# -*- coding: utf-8 -*-
"""
General utility functions to accompany vmod
"""
import numpy as np
import numpy.ma as ma
import time
import math
import os
import subprocess
import utm
from matplotlib.widgets import RectangleSelector
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
import h5py
from IPython.display import Markdown
from IPython.display import display
import matplotlib
from scipy import stats
from scipy.interpolate import interp1d
from skimage.restoration import denoise_nl_means, estimate_sigma


def cart2pol(x1,x2):
    """
    Converts cartesian coordinates to polar coordinates
    
    Parameters:
        x1 (float): x-coordinate (m)
        x2 (float): y-coordinate (m)
        
    Returns:
        theta (float): polar angle (radians)
        r (float): polar radius (m)
    """
    #theta = np.arctan(x2/x1)
    theta = np.arctan2(x2,x1) #sign matters -SH
    r = np.sqrt(x1**2 + x2**2)
    return theta, r

def pol2cart(theta,r):
    """
    Converts polar coordinates to cartesian coordinates
    
    Parameters:
        theta (float): polar angle (radians)
        r (float): polar radius (m)
        
    Returns:
        x1 (float): x-coordinate (m)
        x2 (float): y-coordinate (m)
    """
    x1 = r * np.cos(theta)
    x2 = r * np.sin(theta)
    return x1,x2
    
def ll2utm(lons,lats,z1=None,z2=None):
    """
    Projects lon/lat coordinates using certain utm zone
    
    Parameters:
        lons (array): longitudes to be projected
        lats (array): latitudes to be projected
        z1 (str): utm zone number (optional)
        z2 (str): utm zone letter (optional)
    
    Returns:
        xs (array): projected x-coordinates
        ys (array): projected y-coordinates
        z1 (str): utm zone number
        z2 (str): utm zone letter
    """
    xs,ys=[],[]
    for i in range(len(lats)):
        if z1 is None:
            x,y,z1,z2=utm.from_latlon(lats[i], lons[i])
        else:
            x,y,z1,z2=utm.from_latlon(lats[i], lons[i],force_zone_number=z1, force_zone_letter=z2)
        xs.append(x)
        ys.append(y)
    xs=np.array(xs)
    ys=np.array(ys)
    return xs,ys,z1,z2

def utm2ll(xs,ys,z1,z2):
    """
    Converts utm coordinates into lon/lat
    
    Parameters:
        xs (array): projected x-coordinates to be converted
        ys (array): projected y-coordinates to be converted
        z1 (str): utm zone number
        z2 (str): utm zone letter
    
    Returns:
        lons (array): converted longitudes
        lats (array): converted latitudes
    """
    lons,lats=[],[]
    for i in range(len(xs)):
        lat,lon=utm.to_latlon(xs[i], ys[i], z1, z2)
        lons.append(lon)
        lats.append(lat)
    lons=np.array(lons)
    lats=np.array(lats)
    return lons,lats

def get_quadtree(ref,az,lk,name='quadtree.txt',th=None):
    """
    Downsample an image that is inside a Ref_Selector object
    
    Parameters:
        ref (Ref_Selector): Ref_Selector object that contains the image to be downsampled
        az (array or float): interferogram azimuth angle clockwise from north (degrees)
        lk (array or float): interferogram incidence angle (degrees)
        name (str): csv filename for the output
        th (float): variance threshold
    """
    im=ref.dataset
    if th is None:
        th=np.nanvar(im)/50
    
    quadtree_var(im,az,lk,ref.extent,th,name)

def rewrite_csv(los,ref,old,name='output.txt'):
    """
    Rewrites a csv file that contains the downsampled
    interferogram and has changed the reference
    
    Parameters:
        los (array): new line of sight data
        ref (array): longitude and latitude for the reference pixel (degrees)
        old (str): filename of the csvfile
        name (str): new csv filename for the output
    """
    archivo=open(old,'r')
    lines=archivo.readlines()
    archivo.close()
    
    result=open(name,'w')
    result.write(lines[0].split(':')[0]+': '+str(ref[0])+','+str(ref[1])+', Dimensions:'+lines[0].split('Dimensions:')[1])
    for i in range(len(lines)-1):
        i+=1
        linef=lines[i].split()
        line="%6.3f %6.3f %1.6f %1.6f %1.6f %1.9f %5.0f %5.0f %5.0f %5.0f\n"\
                % (float(linef[0]),float(linef[1]),float(linef[2]),float(linef[3]),los[i-1],float(linef[5]),float(linef[6]),float(linef[7]),float(linef[8]),float(linef[9]))
        
        result.write(line)
    result.close()

def quadtree_var(im,az,inc,extent,th,name='quadtree.txt',ref=None,denoise=True):
    """
    Downsample an image (im) acoording to a variance threshold (th)
    
    Parameters:
        im (array): image (represented by a matrix) to be downsampled
        az (array or float): interferogram azimuth angle clockwise from north (degrees)
        inc (array or float): interferogram incidence angle (degrees)
        extent (array): array with the extent of the image (min_lon,max_lon,min_lat,max_lat)
        th (float): variance threshold
        name (str): filename for the output
        ref (array): reference pixel (lon,lat)
        denoise (boolean): if True a non-local means filter will be applied to the image
    """
    patch_kw = dict(patch_size=5,      # 5x5 patches
                    patch_distance=15,  # 13x13 search area
                    multichannel=True)

    imcp=np.copy(im)
    imcp[np.isnan(im)]=0
    
    
    if denoise:
        imfil=denoise_nl_means(imcp, h=0.6 * 0.5, sigma=0.5, fast_mode=True, **patch_kw)
    else:
        imfil=np.copy(imcp)

    imfil[imcp==0]=np.nan
    
    
    
    fverts=[]
    pointsx=[]
    pointsy=[]
    
    
    def quadtree_im(im,inverts,th):
        """
        Recurrent split function, if the image has a variance lower than th
        the recurrence will stop
        
        Parameters:
            im (array): image to be downsampled
            inverts (array): upper left coordinates (row,col)
            th (float): variance threshold
        """
        #if im.shape[0]*im.shape[1]<=10 or np.sum(np.isnan(im))>=0.9*im.shape[0]*im.shape[1] or np.nanvar(im)<=th:
        #if np.sum(np.isnan(im))>=0.9*im.shape[0]*im.shape[1] or np.nanvar(im)<=th:
        #if np.sum(np.isnan(im))<=10 or np.sum(np.isnan(im))>=0.9*im.shape[0]*im.shape[1] or np.nanvar(im)<=th:
        #if im.shape[0]*im.shape[1]<=10 or np.sum(np.isnan(im))>=0.9*im.shape[0]*im.shape[1] or np.nanvar(im)<=th:
        if im.shape[0]*im.shape[1]<=10 or np.sum(np.logical_not(np.isnan(im)))<2 or np.nanvar(im)<=th:
            y=np.arange(im.shape[0]).astype(float)+inverts[0]
            x=np.arange(im.shape[1]).astype(float)+inverts[1]
            xx,yy=np.meshgrid(x,y)
            xx[np.isnan(im)]=np.nan
            yy[np.isnan(im)]=np.nan
            if not np.isnan(np.nanmean(xx)):
                pointsx.append(int(np.nanmean(xx)))
                pointsy.append(int(np.nanmean(yy)))
                fverts.append((inverts[0],inverts[0]+im.shape[0],inverts[1],inverts[1]+im.shape[1]))

        else:
            halfr=int(im.shape[0]/2)
            halfc=int(im.shape[1]/2)
            quadtree_im(im[0:halfr,0:halfc],inverts,th)
            quadtree_im(im[halfr::,0:halfc],[inverts[0]+halfr,inverts[1]],th)
            quadtree_im(im[0:halfr,halfc::],[inverts[0],inverts[1]+halfc],th)
            quadtree_im(im[halfr::,halfc::],[inverts[0]+halfr,inverts[1]+halfc],th)
            
    quadtree_im(imfil,[0,0],th)
    
    print('Final samples: ',len(fverts))
    ar=open(name,'w')
    if ref is None:
        ar.write('%Reference (Lon,Lat): None, Dimensions: '+str(im.shape[0])+','+str(im.shape[1])+', Extent: '+','.join(str(v) for v in extent)+'\n')
    else:
        ar.write('%Reference (Lon,Lat): '+str(ref[0])+','+str(ref[1])+', Dimensions: '+str(im.shape[0])+','+str(im.shape[1])+', Extent: '+','.join(extent)+'\n')
    xcoords=np.linspace(extent[0],extent[1],im.shape[1])
    ycoords=np.linspace(extent[2],extent[3],im.shape[0])[::-1]
    for i,fvert in enumerate(fverts):
        mean=np.nanmean(im[fvert[0]:fvert[1],fvert[2]:fvert[3]])
        azmean=np.nanmean(az[fvert[0]:fvert[1],fvert[2]:fvert[3]])
        incmean=np.nanmean(inc[fvert[0]:fvert[1],fvert[2]:fvert[3]])
        std=calc_std(im,fvert)
        line="%6.3f %6.3f %1.6f %1.6f %1.6f %1.9f %5.0f %5.0f %5.0f %5.0f\n"\
                % (xcoords[pointsx[i]],ycoords[pointsy[i]],azmean,incmean,mean,std,fvert[0],fvert[1],fvert[2],fvert[3])
        ar.write(line)
    ar.close()

def calc_std(mat,verts):
    """
    Calculates the standard deviation within a sub-matrix. If the std is 0, 
    it makes the sub-matrix bigger
    
    Parameters:
        mat (array): input matrix
        verts (array): indexes to define the sub-matrix (min_row, max_row, min_col, max_col)
        
    Returns:
        std (float): standard deviation of the submatrix
    """
    std = np.nanstd(mat[verts[0]:verts[1],verts[2]:verts[3]]) / np.sum(np.logical_not(np.isnan(mat[verts[0]:verts[1],verts[2]:verts[3]])))
    
    if not std==0:
        return std
    
    if verts[0]==0 and not verts[2]==0:
        newverts=[verts[0],verts[1]+2,verts[2]-1,verts[3]+1]
    elif verts[0]==0 and verts[2]==0:
        newverts=[verts[0],verts[1]+2,verts[2],verts[3]+2]
    elif verts[2]==0 and not verts[0]==0:
        newverts=[verts[0]-1,verts[1]+1,verts[2],verts[3]+2]
    elif verts[1]==mat.shape[0]-1 and not verts[3]==mat.shape[1]-1:
        newverts=[verts[0]-2,verts[1],verts[2]-1,verts[3]+1]
    elif verts[1]==mat.shape[0]-1 and verts[3]==mat.shape[1]-1:
        newverts=[verts[0]-2,verts[1],verts[2]-2,verts[3]]
    elif verts[3]==mat.shape[1]-1 and not verts[1]==mat.shape[0]-1:
        newverts=[verts[0]-1,verts[1]+1,verts[2]-2,verts[3]]
    else:
        newverts=[verts[0]-1,verts[1]+1,verts[2]-1,verts[3]+1]
        
    return calc_std(mat,newverts)
    
def points2map(xs,ys,data):
    """
    Converts a list of points into a matrix
    
    Parameters:
        xs (array): x-coordinates or longitudes 
        ys (array): y-coordinates or latitudes
        data (array): value for each coordinate
    
    Returns:
        qmap (array): matrix with the data
        extent (array): extent of the matrix or image (min_x,max_x,min_y,max_y)
    """
    uxs=np.array(list(set(xs)))
    uys=np.array(list(set(ys)))
    data=np.array(data)
    xint=min_distance(uxs)
    yint=min_distance(uys)
    xnum=int((np.max(xs)-np.min(xs))/xint)
    ynum=int((np.max(ys)-np.min(ys))/yint)
    
    extent=[np.min(xs),np.max(xs),np.min(ys),np.max(ys)]
    
    qmap=np.zeros((ynum,xnum))
    
    lons=np.linspace(np.min(xs),np.max(xs),xnum)
    lats=np.linspace(np.min(ys),np.max(ys),ynum)[::-1]
    
    for i,lon in enumerate(lons):
        for j,lat in enumerate(lats):
            #minn=np.argmin((xs-lon)**2+(ys-lat)**2)
            qmap[j,i]=data[i*len(lons)+j]
    return qmap,extent

def get_defmap(quadfile='quadtree.txt',mask=None, trans=False,cref=True):
    """
    Converts a downsampled image (represented by a text file) into a matrix
    
    Parameters:
        quadfile (str): path for the downsampled image 
        mask (array): boolean matrix for nan values
        trans (array): if True transforms lon/lat into planar coordinates
        cref (boolean): if True reference the matrix with respect to the reference pixel,
        if None reference the matrix with respect to closest value to 0
    
    Returns:
        qmap (array): matrix with the data
        ext (array): extent of the matrix or image (min_x,max_x,min_y,max_y)
        rcoords (array): coordinates for the reference (x-coord,y-coord)
    """
    quad=open(quadfile)
    linesor=quad.readlines()
    quad.close()
    
    dim=(int(linesor[0].split(':')[2].split(',')[0]),int(linesor[0].split(':')[2].split(',')[1]))
    ext=(float(linesor[0].split(':')[3].split(',')[0]), float(linesor[0].split(':')[3].split(',')[1]), float(linesor[0].split(':')[3].split(',')[2]), float(linesor[0].split(':')[3].split(',')[3]))
    rcoords=None
    
    if not 'None' in linesor[0]:
        rcoords=[float(linesor[0].split(':')[1].split(',')[0]),float(linesor[0].split(':')[1].split(',')[1])]
        intc=(ext[1]-ext[0])/dim[1]
        intr=(ext[3]-ext[2])/dim[0]
        col=int((rcoords[0]-ext[0])/intc)
        row=int((ext[3]-rcoords[1])/intr)
    
    lines=[line for line in linesor if not line[0]=='%']
    
    quad=np.zeros(dim)
    quad[:,:]=np.nan
    
    xs,ys,qlos=[],[],[]
    for i,line in enumerate(lines):
        xs.append(float(line.split()[0]))
        ys.append(float(line.split()[1]))
        qlos.append(float(line.split()[4]))
        vert=[int(line.split()[6]),int(line.split()[7]),int(line.split()[8]),int(line.split()[9])]
        quad[vert[0]:vert[1],vert[2]:vert[3]]=float(line.split()[4])
        if rcoords is not None:
            if vert[0]<=row<=vert[1] and vert[2]<=col<=vert[3]:
                posmin=i
        
    xs=np.array(xs)
    ys=np.array(ys)
    uxs=np.array(list(set(xs)))
    uys=np.array(list(set(ys)))
    qlos=np.array(qlos)
    
    if 'None' in linesor[0]:
        posmin=np.argmin(np.abs(qlos))
        rcoords=[xs[posmin],ys[posmin]]
    
    if cref:
        quad[:,:]-=qlos[posmin]
    if mask is not None:
        quad[mask]=np.nan
    
    if trans:
        utmxs,utmys,z1s,z2s=ll2utm([ext[0],ext[1]],[ext[2],ext[3]])
        refxs,refys,z1s,z2s=ll2utm([rcoords[0]],[rcoords[1]])
        ext=[utmxs[0],utmxs[1],utmys[0],utmys[1]]
        rcoords=[refxs[0],refys[0]]
        
    return quad,ext,rcoords

def min_distance(points):
    """
    Gives the minimum distance in a list of coordinates
    
    Parameters:
        points (array): list of points
        
    Returns:
        mini: minimum distance in the list of points
    """
    if len(points)<=3:
        mini=np.inf
        for i in range(len(points)):
            for j in range(i+1,len(points)):
                if np.abs(points[i]-points[j])<mini:
                    mini=np.abs(points[i]-points[j])
        return mini
    
    spoints=np.array(sorted(np.copy(points).tolist()))
    mid=len(points)//2
    midpoint=spoints[mid]
    left=spoints[:mid]
    right=spoints[mid::]
    
    minleft=min_distance(left)
    minright=min_distance(right)
    
    if minleft<minright:
        return minleft
    else:
        return minright

def ll2rc(lon,lat,extent,dims):
    """
    Gives the row and column that correspond to a lon/lat coordinate
    
    Parameters:
        lon (float): longitude coordinate
        lat (float): latitude coordinate
        extent (array): extent of the image (min_lon,max_lon,min_lat,max_lat)
        dims (array): dimensions of the image (rows,columns)
        
    Returns:
        row (int): row that corresponds to the coordinate
        col (int): column that corresponds to the coordinate
    """
    lonr1, lonr2, latr1, latr2=extent
    lons=np.linspace(lonr1,lonr2,dims[1])
    lats=np.linspace(latr1,latr2,dims[0])[::-1]

    row=np.argmin(np.abs(lats-lat))

    col=np.argmin(np.abs(lons-lon))
    
    return row,col

def get_closest_point(row,col,dataset):
    """
    Gets the closest point for a pixel that is not nan
    
    Parameters:
        row (int): row for the pixel
        col (int): column for the pixel
        dataset (array): matrix that represents an image
    
    Returns:
        row (int): row for the closest pixel that is not nan
        col (int): column for the closest pixel that is not nan
    """
    if not np.isnan(dataset[row,col]):
        return row,col
    else:
        x=np.linspace(0,dataset.shape[1],dataset.shape[1])
        y=np.linspace(0,dataset.shape[0],dataset.shape[0])
        XX,YY=np.meshgrid(x,y)
        
        XX[np.isnan(dataset)]=np.nan
        YY[np.isnan(dataset)]=np.nan
        
        dif=np.sqrt((XX-col)**2+(YY-row)**2)
        
        row,col=np.unravel_index(np.nanargmin(dif), dif.shape)
        return row,col
    
def read_dataset_h5(h5file,key,plot=True,aoi=None):
    """
    Reads and plots the output from mintpy 
    
    Parameters:
        h5file (str): path to h5 file
        key (str): key for certain dataset (e.g., coherence)
        plot (boolean): if True plots the dataset
        aoi (AOI_Selector): AOI_Selector to get the area of interest in the dataset
    
    Returns:
        dataset (array): matrix that represents the dataset
    """
    h5f=h5py.File(h5file)
    dataset=h5f[key][:]
    
    lons=[float(h5f.attrs['LON_REF1']), float(h5f.attrs['LON_REF2'])]
    lats=[float(h5f.attrs['LAT_REF2']), float(h5f.attrs['LAT_REF3'])]
    
    lonr1,lonr2,latr1,latr2=np.min(lons),np.max(lons),np.min(lats),np.max(lats)
    #lonr1, lonr2, latr1, latr2 = float(h5f.attrs['LON_REF1']), float(h5f.attrs['LON_REF2']), float(h5f.attrs['LAT_REF2']), float(h5f.attrs['LAT_REF3'])
    
    extent=[lonr1,lonr2,latr1,latr2]
    h5f.close()
    
    if aoi is not None:
        row1,col1=ll2rc(aoi.x1,aoi.y2,extent,dataset.shape)
        row2,col2=ll2rc(aoi.x2,aoi.y1,extent,dataset.shape)
        extent=[aoi.x1,aoi.x2,aoi.y1,aoi.y2]
        dataset=dataset[row1:row2,col1:col2]
    
    vmin = np.nanpercentile(dataset, 1)
    vmax = np.nanpercentile(dataset, 99)
    
    fig, ax = plt.subplots()
        
    fig.suptitle(key, fontsize=16)
    
    if 'coherence' in key.lower():
        cmap = matplotlib.cm.gist_gray.copy()
        cmap.set_bad('black')
        im=ax.imshow(dataset, cmap=cmap, extent=extent, vmin=vmin, vmax=vmax)
    else:
        cmap=plt.cm.jet
        im=ax.imshow(dataset,cmap=cmap,extent=extent,vmin=vmin, vmax=vmax)
    ax.set_ylabel('Latitude (°)')
    ax.set_xlabel('Longitude (°)')
    plt.colorbar(im,orientation='horizontal')
    print(extent)
    return dataset

def read_gnss_csv(csvfile,trans=False):
    """
    Reads csv file with a gnss dataset 
    
    Parameters:
        csvfile (str): path to csv file
        trans (boolean): projects the lon/lat coordinates into plane coordinates
        
    Returns:
        names (array): names of the stations
        xs/lons (array): x-coordinates or longitudes for the stations
        ys/lats (array): y-coordinates or latitudes for the stations
        uxs (array): deformation in the east component
        uys (array): deformation in the north component
        uzs (array): deformation in the vertical component
        euxs (array): uncertainties in the deformation in the east component
        euys (array): uncertainties in the deformation in the north component
        euzs (array): uncertainties in the deformation in the vertical component
        ref (array): origin coordinate if a projection was made
    """
    archivo=open(csvfile,'r')
    linesor=archivo.readlines()
    archivo.close()
    
    lines=[line for line in linesor if not '%' in line]
    
    names,lons,lats,uxs,uys,uzs,euxs,euys,euzs=[],[],[],[],[],[],[],[],[]
    for line in lines:
        names.append(line.split()[0])
        lons.append(float(line.split()[1]))
        lats.append(float(line.split()[2]))
        uxs.append(float(line.split()[3]))
        uys.append(float(line.split()[4]))
        uzs.append(float(line.split()[5]))
        euxs.append(float(line.split()[6]))
        euys.append(float(line.split()[7]))
        euzs.append(float(line.split()[8]))
        
    lons=np.array(lons)
    lots=np.array(lats)
    uxs=np.array(uxs)
    uys=np.array(uys)
    uzs=np.array(uzs)
    euxs=np.array(euxs)
    euys=np.array(euys)
    euzs=np.array(euzs)
    
    if trans:
        xs,ys,z1s,z2s=ll2utm(lons,lats)
        meanx=np.mean(xs)
        meany=np.mean(ys)
        xs-=meanx
        ys-=meany
        ref=[meanx,meany,z1s,z2s]
        return names,xs,ys,uxs,uys,uzs,euxs,euys,euzs,ref
    else:
        return names,np.array(lons),np.array(lats),uxs,uys,uzs,euxs,euys,euzs

def read_insar_csv(csvfile,trans=False,unit='m',ori=None,cref=True):
    """
    Reads csv file with a downsampled InSAR dataset 
    
    Parameters:
        csvfile (str): path to csv file
        trans (boolean): if True projects the lon/lat coordinates into plane coordinates (default False)
        unit (str): units for the coordinates (m for meters or deg for degrees) (default m)
        ori (array): coordinates (lon,lat) that will be the origin for the projection
        cref
        
    Returns:
        names (array): names of the stations
        xs/lons (array): x-coordinates or longitudes for the stations
        ys/lats (array): y-coordinates or latitudes for the stations
        uxs (array): deformation in the east component
        uys (array): deformation in the north component
        uzs (array): deformation in the vertical component
        euxs (array): uncertainties in the deformation in the east component
        euys (array): uncertainties in the deformation in the north component
        euzs (array): uncertainties in the deformation in the vertical component
        ref (array): origin coordinate if a projection was made
    """
    archivo=open(csvfile,'r')
    linesor=archivo.readlines()
    archivo.close()
    
    lines=[line for line in linesor if not '%' in line]
    
    lons,lats,azs,lks,los,elos=[],[],[],[],[],[]
    for line in lines:
        lons.append(float(line.split()[0]))
        lats.append(float(line.split()[1]))
        azs.append(float(line.split()[2]))
        lks.append(float(line.split()[3]))
        los.append(float(line.split()[4]))
        elos.append(float(line.split()[5]))
    lons=np.array(lons)
    lats=np.array(lats)
    azs=np.array(azs)
    lks=np.array(lks)
    los=np.array(los)
    elos=np.array(elos)
    
    if linesor[0][0]=='%' and not 'None' in linesor[0]:
        ref=[float(linesor[0].split(':')[1].split(',')[0]),float(linesor[0].split(':')[1].split(',')[1])]
        posmin=np.argmin((lons-ref[0])**2+(lats-ref[1])**2)
    elif linesor[0][0]=='%' and cref:
        posmin=np.argmin(np.abs(los))
        ref=[lons[posmin],lats[posmin]]
        los-=los[posmin]
    else:
        ref=None
    if trans and unit=='m':
        xs,ys,z1s,z2s=ll2utm(lons,lats)
        xref,yref,z1s,z2s=ll2utm([ref[0]],[ref[1]])
        meanx=np.mean(xs)
        meany=np.mean(ys)
        xs-=meanx
        ys-=meany
        ref=[xref-meanx,yref-meany,meanx,meany,str(z1s)+str(z2s)]
        return xs,ys,azs,lks,los,elos,ref
    else:
        return lons,lats,azs,lks,los,elos,ref

def plot_gnss(xs,ys,uxs,uys,uzs,title=None,names=None,euxs=None,euys=None,euzs=None,scl=None,unit='m',figsize=None):
    """
    Plots gnss dataset, horizontal velocities are represented by red arrows
    vertical velocities are represented by black lines
    
    Parameters:
        xs/lons (array): x-coordinates or longitudes for the stations (m/deg)
        ys/lats (array): y-coordinates or latitudes for the stations (m/deg)
        uxs (array): deformation in the east component
        uys (array): deformation in the north component
        uzs (array): deformation in the vertical component
        title (str): title for the plot
        names (array): names for the stations
        euxs (array): uncertainties in the deformation in the east component
        euys (array): uncertainties in the deformation in the north component
        euzs (array): uncertainties in the deformation in the vertical component
        scl (float): scale for the arrows
        unit (str): unit for the coordinates (m for meters or deg for degrees)
        figsize (tuple): size for the matplotlib figure
    """
    #Plotting GPS deformation
    if unit=='m':
        norm=1e3
    else:
        norm=1.0
    ratio=0.2
    if figsize:
        plt.figure(figsize=figsize)
    else:
        plt.figure(figsize=(5,5))
    
    if title:
        plt.title(title)
    else:
        plt.title('GPS data')
    plt.scatter(xs/norm,ys/norm,s=10)
    hmax=np.max(np.sqrt(uxs**2+uys**2))
    extentx=(np.max(xs)-np.min(xs))/norm
    extenty=(np.max(ys)-np.min(ys))/norm
    if unit=='m':
        limsx=[-extentx/2-ratio*extentx,extentx/2+ratio*extentx]
        limsy=[-extenty/2-ratio*extenty,extenty/2+ratio*extenty]
    else:
        limsx=[np.min(xs)-2.25*ratio*extentx,np.max(xs)+2.25*ratio*extentx]
        limsy=[np.min(ys)-2.25*ratio*extenty,np.max(ys)+2.25*ratio*extenty]
    extent=np.max([limsx[1]-limsx[0],limsy[1]-limsy[0]])

    if unit=='m':
        sposx=limsx[0]+ratio/4*(limsx[1]-limsx[0])
        sposy=limsy[1]-ratio/4*(limsy[1]-limsy[0])
        sposy2=limsy[0]+ratio/4*(limsx[1]-limsx[0])
    else:
        sposx=limsx[0]+ratio/10*(limsx[1]-limsx[0])
        sposy=limsy[1]-ratio/10*(limsy[1]-limsy[0])
        sposy2=limsy[0]+ratio/10*(limsx[1]-limsx[0])-0.05*(limsy[1]-limsy[0])
    sposy1=sposy-0.05*(limsy[1]-limsy[0])
    
    
    if scl is None:
        scale=ratio*extent/hmax
        sc=hmax/2
        scl=round(sc*100,2)
    else:
        sc=scl/100
        hmax=2*sc
        scale=ratio*extent/hmax
    logsc=scale/10**int(np.log10(scale))
    ax = plt.gca()
    for i in range(len(xs)):
        plt.annotate("", xy=(xs[i]/norm+uxs[i]*scale, ys[i]/norm+uys[i]*scale), xytext=(xs[i]/norm, ys[i]/norm),arrowprops=dict(arrowstyle="->",color="red"))
        if not (euxs is None and euys is None):
            ax.add_patch(Ellipse(xy=(xs[i]/norm+uxs[i]*scale, ys[i]/norm+uys[i]*scale), width=euxs[i]*scale*2, height=euys[i]*scale*2, color="grey", fill=False, lw=2))
        plt.annotate("", xy=(xs[i]/norm, ys[i]/norm+uzs[i]*scale), xytext=(xs[i]/norm, ys[i]/norm),arrowprops=dict(arrowstyle="-",color="black"))
        if names is not None:
            plt.annotate(names[i], xy=(xs[i]/norm, ys[i]/norm-0.05*(limsy[1]-limsy[0])), xytext=(xs[i]/norm, ys[i]/norm-0.05*(limsy[1]-limsy[0])),color='blue')
    if euzs is not None:
        plt.errorbar(xs/norm,ys/norm+uzs*scale,euzs*scale,fmt='bo',ms=1)
    plt.annotate("", xy=(sposx+sc*scale, sposy), xytext=(sposx, sposy),arrowprops=dict(arrowstyle="->",color="red"))
    plt.annotate(str(scl)+r"cm/yr", xy=(sposx, sposy1), xytext=(sposx, sposy1),color='red')
    
    plt.annotate("", xy=(sposx, sposy2+sc*scale), xytext=(sposx, sposy2),arrowprops=dict(arrowstyle="-",color="black"))
    plt.annotate(str(scl)+r"cm/yr", xy=(sposx, sposy2), xytext=(sposx, sposy2),color='black')
    if unit=='m':
        plt.ylabel('Y(km)')
        plt.xlabel('X(km)')
    else:
        plt.ylabel('Lat(°)')
        plt.xlabel('Lon(°)')
    
    plt.xlim(limsx)
    plt.ylim(limsy)
    ax = plt.gca()
    ax.xaxis.set_major_locator(plt.MaxNLocator(4))
    #plt.axis('square')
    #plt.axis('equal')
    #plt.axis('scaled')
    plt.show()

def los2npy(los,quadfile,maskfile=None,output=None,cref=False):
    """
    Creates matrix representing downsampled InSAR dataset replacing los deformation
    and save it in a npy file
    
    Parameters:
        los (array): new line of sight deformation dataset
        quadfile (array): downsampled InSAR dataset
        maskfile (str): npy file with boolean matrix that indicates nan values
        output (str): npy filename for the output
        cref (boolean): if True reference the matrix with respect to the reference pixel,
        if None reference the matrix with respect to closest value to 0
        
    Returns:
        qmap (array): saved matrix that represents downsampled InSAR dataset
        extent (array): coordinates with the extent of the image
    """
    archivo=open(quadfile,'r')
    lines=archivo.readlines()
    archivo.close()

    dim=[int(lines[0].split('Dimensions:')[1].split(',')[i]) for i in range(2)]
    
    result=open('temp.txt','w')
    result.write(lines[0])
    for i in range(len(lines)-1):
        i+=1
        linef=lines[i].split()
        line="%6.3f %6.3f %1.6f %1.6f %1.6f %1.9f %5.0f %5.0f %5.0f %5.0f\n"\
                    % (float(linef[0]),float(linef[1]),float(linef[2]),float(linef[3]),los[i-1],float(linef[5]),float(linef[6]),float(linef[7]),float(linef[8]),float(linef[9]))

        result.write(line)
    result.close()
    
    if maskfile:
        mask_des=np.load(maskfile)
    else:
        mask_des=np.zeros((dim[0],dim[1]))
        mask_des=mask_des>0
        
    qmap,extent,rcoords=get_defmap('temp.txt',mask=mask_des,trans=False,cref=cref)

    subprocess.call('rm -rf temp.txt',shell=True)
    
    if output:
        np.save(output,qmap)
    
    return qmap,extent
    
class AOI_Selector:
    """
    Class to create an interactive tool to select the area of interest from
    a mintpy output h5 file
    
    Attributes:
        wl (float): wavelength of the mission
        coh (array): coherence dataset in h5 file
        cohth (float): the pixels that have a coherence value below this won't be plot
        image (array): dataset to be plot
        extent (array): coordinates for the extent of the image (min_lon,max_lon,min_lat,max_lat)
        x1 (float): upper left x-coordinate or longitude of area of interest
        y1 (float): upper left y-coordinate or latitude of area of interest
        x2 (float): lower right x-coordinate or longitude of area of interest
        y2 (float): lower right y-coordinate or latitude of area of interest
        vmin (float): minimum value for colorbar
        vmax (float): maximum value for colorbar
        fig (matplotlib Figure): figure that plots the dataset
        current_ax (matplotlib subplot): subplot that has the plot for the dataset
    """
    def __init__(self,
                 h5file,
                 key,
                 coh=None,cohth=0,
                 fig_xsize=None, fig_ysize=None,
                 cmap=plt.cm.gist_gray,
                 vmin=None, vmax=None,
                 drawtype='box'
                ):
        
        h5f=h5py.File(h5file)
        keys=[ke for ke in h5f.keys()]
        if 'timeseries' in keys:
            try:
                dates=np.array([h5f['date'][:][i].decode('utf-8') for i in range(len(h5f['date'][:]))])
                print('The possible dates are:',dates)
                timeseries=h5f['timeseries'][:]
                if key in dates:
                    velocity=timeseries[dates==key][0]
                else:
                    print('The date was not found in this dataset')
                    velocity=timeseries[-1]
            except:
                raise Exception('The dataset does not exist in this file')
        else:
            if key in keys:
                velocity=h5f[key][:]
            else:
                try:
                    velocity=h5f['velocity'][:]
                except:
                    raise Exception('This dataset does not have LOS deformation')
                    
        lons=[float(h5f.attrs['LON_REF1']), float(h5f.attrs['LON_REF2'])]
        lats=[float(h5f.attrs['LAT_REF2']), float(h5f.attrs['LAT_REF3'])]
        lonr1,lonr2,latr1,latr2=np.min(lons),np.max(lons),np.min(lats),np.max(lats)
        #lonr1, lonr2, latr1, latr2 = float(h5f.attrs['LON_REF1']), float(h5f.attrs['LON_REF2']), float(h5f.attrs['LAT_REF2']), float(h5f.attrs['LAT_REF3'])
        self.wl=float(h5f.attrs['WAVELENGTH'])
        
        h5f.close()

        velocity[velocity==0]=np.nan
        
        if coh is not None:
            self.coh=coh
            self.cohth=cohth
            velocity[coh<cohth]=np.nan
        
        print('Please select the area of interest')
        
        self.image = velocity
        self.extent=[lonr1,lonr2,latr1,latr2]
        self.x1 = None
        self.y1 = None
        self.x2 = None
        self.y2 = None
        if not vmin:
            self.vmin = -np.nanmax(np.abs(velocity))
        else:
            self.vmin = vmin
        if not vmax:
            self.vmax = np.nanmax(np.abs(velocity))
        else:
            self.vmax = vmax
        if fig_xsize and fig_ysize:
            self.fig, self.current_ax = plt.subplots(figsize=(fig_xsize, fig_ysize))
        else:
            self.fig, self.current_ax = plt.subplots()
        self.fig.suptitle('Area-Of-Interest Selector', fontsize=16)
        
        if 'coherence' in key.lower():
            self.cmap = matplotlib.cm.gist_gray.copy()
            cmap.set_bad('black')
            im=self.current_ax.imshow(self.image, cmap=self.cmap, extent=self.extent, vmin=self.vmin, vmax=self.vmax)
        else:
            self.cmap=plt.cm.jet
            im=self.current_ax.imshow(self.image, cmap=plt.cm.jet, extent=self.extent, vmin=self.vmin, vmax=self.vmax)
        
        self.current_ax.set_ylabel('Latitude (°)')
        self.current_ax.set_xlabel('Longitude (°)')
        plt.colorbar(im,orientation='horizontal')

        def toggle_selector(self, event):
            """
            Activates the selector tool in matplotlib
            """
            print(' Key pressed.')
            if event.key in ['Q', 'q'] and toggle_selector.RS.active:
                print(' RectangleSelector deactivated.')
                toggle_selector.RS.set_active(False)
            if event.key in ['A', 'a'] and not toggle_selector.RS.active:
                print(' RectangleSelector activated.')
                toggle_selector.RS.set_active(True)

        toggle_selector.RS = RectangleSelector(self.current_ax, self.line_select_callback,
                                               drawtype=drawtype, useblit=True,
                                               button=[1, 3],  # don't use middle button
                                               minspanx=0, minspany=0,
                                               spancoords='pixels',
                                               rectprops = dict(facecolor='red', edgecolor = 'yellow',
                                                                alpha=0.3, fill=True),
                                               interactive=True)
        plt.connect('key_press_event', toggle_selector)

    def line_select_callback(self, eclick, erelease):
        """
        Catch the coordinates from the selector tool
        """
        'eclick and erelease are the press and release events'
        self.x1, self.y1 = eclick.xdata, eclick.ydata
        self.x2, self.y2 = erelease.xdata, erelease.ydata
        print("(%3.2f, %3.2f) --> (%3.2f, %3.2f)" % (self.x1, self.y1, self.x2, self.y2))
        print(" The button you used were: %s %s" % (eclick.button, erelease.button))
    
class Ref_Insar_Selector_Pre:
    """
    Class to create an interactive tool to select a reference pixel from 
    an AOI_Selector object
    
    Attributes:
        wl (float): wavelength of the mission
        xref (float): x-coordinate for the reference
        yref (float): y-coordinate for the reference
        coh (array): coherence dataset in h5 file
        dataset (array): dataset to be plot
        extent (array): coordinates for the extent of the image (min_lon,max_lon,min_lat,max_lat)
    """
    def __init__(self,aoi):
        self.xref=None
        self.yref=None
        
        velocity=aoi.image
        extent=aoi.extent
        lonr1,lonr2,latr1,latr2=extent
        self.wl=aoi.wl

        velocity[velocity==0]=np.nan
        if aoi.coh is not None:
            velocity[aoi.coh<aoi.cohth]=np.nan

        lons=np.linspace(lonr1,lonr2,velocity.shape[1])
        lats=np.linspace(latr1,latr2,velocity.shape[0])[::-1]
        
        if aoi.x1 is None:
            row1=0
            row2=velocity.shape[0]
            col1=0
            col2=velocity.shape[1]
            self.extent=extent
        else:
            row1,col1=ll2rc(aoi.x1,aoi.y2,extent,velocity.shape)
            row2,col2=ll2rc(aoi.x2,aoi.y1,extent,velocity.shape)
            self.extent=[aoi.x1,aoi.x2,aoi.y1,aoi.y2]
        
        self.dataset=np.copy(velocity[row1:row2,col1:col2])
        fig, ax = plt.subplots()
        
        fig.suptitle('Reference Selector', fontsize=16)
        
        im=ax.imshow(self.dataset,cmap=aoi.cmap,extent=self.extent,vmin=aoi.vmin, vmax=aoi.vmax)
        line,=ax.plot([], [],'ko')
        
        ax.set_ylabel('Latitude (°)')
        ax.set_xlabel('Longitude (°)')

        plt.colorbar(im,orientation='horizontal')

        def on_click(event):
            """
            Captures the coordinates when a user click on the plot
            """
            row,col=ll2rc(event.xdata,event.ydata,self.extent,self.dataset.shape)
            
            row,col=get_closest_point(row,col,self.dataset)
            line.set_xdata(event.xdata)
            line.set_ydata(event.ydata)
            
            self.xref=event.xdata
            self.yref=event.ydata
            
            self.dataset-=self.dataset[row,col]
            
            im.set_data(self.dataset)

        plt.connect('button_press_event', on_click)

        plt.show()
        
class Ref_Insar_Selector:
    """
    Class to create an interactive tool to select a reference pixel from 
    a csv file that represents a downsampled InSAR dataset
    
    Attributes:
        wl (float): wavelength of the mission
        xref (float): x-coordinate for the reference
        yref (float): y-coordinate for the reference
        auxdata (tuple): auxiliary information (longitudes, latitudes, azimuth angles, incidence angles, uncertainties)
        los (array): list with line of sight deformation
        filename (str): path to the csv file with the downsampled InSAR dataset
        dataset (array): matrix with the line of sight deformation
    """
    def __init__(self,csvfile,mask=None,vmin=None,vmax=None):
        
        velocity,extent,refcoords=get_defmap(csvfile,mask)
        
        lons,lats,azs,lks,los,elos,ref=read_insar_csv(csvfile,unit='deg')
        
        self.xref=ref[0]
        self.yref=ref[1]
        self.extent=extent
        self.auxdata=(lons,lats,azs,lks,elos)
        self.los=np.copy(los)
        self.filename=csvfile
        
        lonr1,lonr2,latr1,latr2=extent

        self.dataset=np.copy(velocity)
        fig, ax = plt.subplots()
        
        fig.suptitle('Reference Selector', fontsize=16)
        
        if vmin is None:
            im=ax.imshow(self.dataset,cmap='jet',extent=self.extent,vmin=-np.nanmax(np.abs(velocity)), vmax=np.nanmax(np.abs(velocity)))
        else:
            im=ax.imshow(self.dataset,cmap='jet',extent=self.extent,vmin=vmin, vmax=vmax)
        #line,=ax.plot([refcoords[0]], [refcoords[1]],'ko',label='Reference')
        line,=ax.plot([ref[0]], [ref[1]],'ko',label='Reference')
        
        ax.set_ylabel('Latitude (°)')
        ax.set_xlabel('Longitude (°)')

        plt.colorbar(im,orientation='horizontal')

        def on_click(event):
            row,col=ll2rc(event.xdata,event.ydata,self.extent,self.dataset.shape)
            
            row,col=get_closest_point(row,col,self.dataset)
            line.set_xdata(event.xdata)
            line.set_ydata(event.ydata)
            
            self.xref=event.xdata
            self.yref=event.ydata
            
            self.dataset-=self.dataset[row,col]
            
            im.set_data(self.dataset)
            
            self.los-=self.dataset[row,col]
            
            rewrite_csv(self.los,[self.xref,self.yref],old=csvfile,name=csvfile.split('.')[0]+'_ref.'+csvfile.split('.')[1])
            
        plt.legend()

        plt.connect('button_press_event', on_click)

        plt.show()
    


import numpy as np
from .. import util
from . import Source
from .okada import Okada
from ..data import Gnss

class Regsill(Source):
    """
    Class to represent a regularized sill using dislocation patches from Okada (1985) model.

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    typ : str
        type of dislocation, open for tensile and slip for fault
    ln : number of segments in the lenght
    wn : number of segments in the width
    xcen : x-coordinate for the center of the sill/fault
    ycen : y-coordinate for the center of the sill/fault
    depth : depth of the center of the sill/fault
    length : length of the sill/fault
    width : width of the sill/fault
    strike : orientation of the sill/fault clockwise from north
    dip : dip angle of the fault 0 horizontal, 90 vertical
    lamb : regularization constant
    """
    def __init__(self, data, typ=None, ln=None,wn=None, xcen=None,ycen=None,depth=None,length=None,width=None,strike=None,dip=None,lamb=None,rake=None):
        self.typ = 'open' if typ == 'open' or typ is None else 'slip'
        self.ln     = 1 if ln     is None else ln
        self.wn     = 1 if wn     is None else wn
        self.xcen   = 0 if xcen   is None else xcen
        self.ycen   = 0 if ycen   is None else ycen
        self.depth  = 1 if depth  is None else depth
        self.length = 0 if length is None else length
        self.width  = 0 if width  is None else width
        self.strike = 0 if strike is None else strike
        self.dip    = 0 if dip    is None else dip
        self.rake   = 0 if rake   is None else rake
        self.lamb   = 0 if lamb   is None else lamb
        
        super().__init__(data)
        
        self.reg    = True
        
    def time_dependent(self):
        return False
    
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=()
        for i in range(self.ln*self.wn):
            if self.typ=='open':
                self.parameters+=('open'+str(i),)
            else:
                self.parameters+=('slip'+str(i),)
    
    def set_x0(self,x0):
        """
        Initial guess for the openings/slips
        """
        self.get_parnames()
        self.x0=[x0]*(self.ln*self.wn)
    
    def set_bounds(self,low_bound,high_bound):
        """
        Bounds for the openings/slips
        """
        self.get_parnames()
        self.low_bounds=[low_bound]*(self.ln*self.wn)
        self.high_bounds=[high_bound]*(self.ln*self.wn)
        
    def rotate_xyz(self):
        """
        Rotation for the patches in the sill/fault
        """
        cx=self.xcen
        cy=self.ycen
        cz=-self.depth
        wp=self.width*np.cos(np.radians(self.dip))
        wr=self.width*np.sin(np.radians(self.dip))
        l=self.length
        phi=self.strike
        x1 = cx + wp/2 * np.cos(np.radians(phi)) - l/2 * np.sin(np.radians(phi))
        y1 = cy + wp/2 * np.sin(np.radians(phi)) + l/2 * np.cos(np.radians(phi))
        z1 = cz - wr/2
        x2 = cx - wp/2 * np.cos(np.radians(phi)) - l/2 * np.sin(np.radians(phi))
        y2 = cy - wp/2 * np.sin(np.radians(phi)) + l/2 * np.cos(np.radians(phi))
        z2 = cz + wr/2
        x3 = cx - wp/2 * np.cos(np.radians(phi)) + l/2 * np.sin(np.radians(phi))
        y3 = cy - wp/2 * np.sin(np.radians(phi)) - l/2 * np.cos(np.radians(phi))
        z3 = cz + wr/2
        x4 = cx + wp/2 * np.cos(np.radians(phi)) + l/2 * np.sin(np.radians(phi))
        y4 = cy + wp/2 * np.sin(np.radians(phi)) - l/2 * np.cos(np.radians(phi))
        z4 = cz - wr/2
        
        return [x1,x2,x3,x4],[y1,y2,y3,y4],[z1,z2,z3,z4]
        
    def get_centers(self):
        """
        Returns the centers for sill/fault patches
        """
        xc=self.xcen
        yc=self.ycen
        zc=-self.depth
        ln=self.ln
        wn=self.wn
        lslice=self.length/self.ln
        wslice=self.width/self.wn
        fwc=self.xcen-self.width/2+self.width/(2*self.wn)
        flc=self.ycen-self.length/2+self.length/(2*self.ln)
        #print(fwc,flc)
        xcs,ycs,zcs=[],[],[]
        if wn%2==0:
            wi=self.wn/2
        else:
            wi=(self.wn-1)/2
            
        if ln%2==0:
            li=self.ln/2
        else:
            li=(self.ln-1)/2
            
        for i in range(int(wi)):
            wfake=2*np.abs(fwc-self.xcen+float(i)*wslice)
            for j in range(int(li)):
                lfake=2*np.abs(flc-self.ycen+float(j)*lslice)
                xs,ys,zs=self.rotate_xyz()
                for x in xs:
                    xcs.append(x)
                for y in ys:
                    ycs.append(y)
                for z in zs:
                    zcs.append(z)
        #print('Puntos 1',len(xcs),wn%2,ln%2)
        if not self.ln%2==0:
            for j in range(int(li)):
                wfake=0
                lfake=2*np.abs(flc-self.ycen+float(j)*lslice)
                xs,ys,zs=self.rotate_xyz()
                for x in xs[1:3]:
                    xcs.append(x)
                for y in ys[1:3]:
                    ycs.append(y)
                for z in zs[1:3]:
                    zcs.append(z)
        #print('Puntos 2',len(xcs))
        if not self.wn%2==0:
            for i in range(int(wi)):
                wfake=2*np.abs(fwc-self.xcen+float(i)*wslice)
                lfake=0
                xs,ys,zs=self.rotate_xyz()
                for x in xs[0:2]:
                    xcs.append(x)
                for y in ys[0:2]:
                    ycs.append(y)
                for z in zs[0:2]:
                    zcs.append(z)
        #print('Puntos 3',len(xcs))
        if (not self.wn%2==0) and (not self.ln%2==0):
            xcs.append(self.xcen)
            ycs.append(self.ycen)
            zcs.append(-self.depth)
        #print('Puntos 4',len(xcs))
        return xcs,ycs,zcs
    
    def get_greens(self):
        """
        Greens functions for the patches
        """
        ln=self.ln
        wn=self.wn
        x=self.data.xs
        y=self.data.ys
        
        dat1 = Gnss()
        dat1.add_xs(x)
        dat1.add_ys(y)
        
        dat1.add_data(x*0,x*0,x*0)
        
        oki=Okada(dat1)
        #print('Tipo',self.typ)
        oki.set_type(self.typ)
        
        xcs,ycs,zcs=self.get_centers()
        xo=[self.xcen,self.ycen,self.depth,self.length,self.width,1,self.strike,self.dip]
        
        defo=oki.forward(xo)
        
        slength=self.length/ln
        swidth=self.width/wn
        
        if self.typ=='open':
            op=1
            sl=0
        else:
            op=0
            sl=1
        G=np.zeros((len(defo),ln*wn))
        for i in range(len(xcs)):
            if self.typ=='open':
                xp=[xcs[i],ycs[i],-zcs[i],slength,swidth,op,self.strike,self.dip]
            else:
                xp=[xcs[i],ycs[i],-zcs[i],slength,swidth,sl,self.strike,self.dip,self.rake]
            defo=oki.forward(xp)
            G[:,i]=defo
        return G
    
    def get_laplacian(self):
        """
        Second order laplacian to regularized the patches in the sill/fault
        """
        ln=self.ln
        wn=self.wn
        xcs,ycs,zcs=self.get_centers()
        L=np.zeros((ln*wn,ln*wn))
        for i in range(len(xcs)):
            dist=(np.array(xcs)-xcs[i])**2+(np.array(ycs)-ycs[i])**2+(np.array(zcs)-zcs[i])**2
            pos=np.argsort(dist)
            if dist[pos[1]]==dist[pos[2]] and dist[pos[1]]==dist[pos[3]] and dist[pos[1]]==dist[pos[4]]:
                L[i,pos[0]]=-2
                L[i,pos[1]]=1
                L[i,pos[2]]=1
                L[i,pos[3]]=1
                L[i,pos[4]]=1
            elif dist[pos[1]]==dist[pos[2]] and dist[pos[1]]==dist[pos[3]]:
                L[i,pos[0]]=-2
                L[i,pos[1]]=1
                L[i,pos[2]]=1
                L[i,pos[3]]=1
            elif dist[pos[1]]==dist[pos[2]]:
                L[i,pos[0]]=-2
                L[i,pos[1]]=1
                L[i,pos[2]]=1
        return L
    
    def model(self,x,y,*ops):
        """
        3d displacement produce by the discretized sill
          
        Parameters:
            ops: array that contains the opening/slip values for the patches
        """
        G=self.get_greens()
        model=np.array(ops)
        data=G@model
        
        ux=data[0:len(x)]
        uy=data[len(x):2*len(x)]
        uz=data[2*len(x):3*len(x)]
        
        return ux,uy,uz
        
    def plot_patches(self,ops):
        """
        Auxiliary function to plot the opening/slip values
        
        Parameters:
            ops: array that contains the opening/slip values for the patches
        """
        import matplotlib
        import matplotlib.pyplot as plt
        import numpy as np
        
        ln=self.ln
        wn=self.wn

        ok1 = Okada(self.data)
        ok1.set_type('open')
        xs,ys,zs=self.get_centers(0,0,0,self.length,self.width,0,0)

        patches=[]
        fig, ax = plt.subplots()
        for i in range(len(xs)):
            rect = matplotlib.patches.Rectangle((xs[i]-self.length/(2*ln),ys[i]-self.width/(2*wn)), self.length/ln, self.width/wn)
            patches.append(rect)

        # values as numpy array
        values = ops

        # define the norm 
        norm = plt.Normalize(values.min(), values.max())
        coll = matplotlib.collections.PatchCollection(patches, cmap='viridis',
                                                      norm=norm, match_original = True)

        coll.set_array(values)
        polys = ax.add_collection(coll)
        fig.colorbar(polys, label='Opening(m)')

        plt.xlim(-self.width/2,self.width/2)
        plt.ylim(-self.length/2,self.length/2)

        plt.show()



from .source import Source
import os
import sys

path = os.path.dirname(os.path.abspath(__file__))

for py in [f[:-3] for f in os.listdir(path) if f.endswith('.py') and f != '__init__.py']:
    mod = __import__('.'.join([__name__, py]), fromlist=[py])
    classes = [getattr(mod, x) for x in dir(mod) if isinstance(getattr(mod, x), type)]
    for cls in classes:
        setattr(sys.modules[__name__], cls.__name__, cls)


import numpy as np
from . import Data
from vmod import util

class Gnss(Data):
    """
    Class that represents the GNSS geodetic datatype
    
    Attributes:
        ux: deformation in the east (m)
        uy: deformation in the north (m)
        uz: deformation in the vertical (m)
        errx: uncertainty in the east deformation (m)
        erry: uncertainty in the north deformation (m)
        errz: uncertainty in the vertical deformation (m)
    """
    def __init__(self):
        self.ux=None
        self.uy=None
        self.uz=None
        self.errx=None
        self.erry=None
        self.errz=None
        super().__init__()    
        
    def add_ux(self,ux):
        """
        Adds the deformation in the east component
        
        Parameters:
            ux (array): deformation in the east component
        """
        self.add_comp(ux,'ux')
        self.ux=ux
    
    def add_errx(self,errx):
        """
        Adds the uncertainty in the deformation for the east component
        
        Parameters:
            errx (array): uncertainty in the east component
        """
        self.add_errcomp(errx,'ux')
        self.errx=errx
    
    def add_uy(self,uy):
        """
        Adds the deformation in the north component
        
        Parameters:
            uy (array): deformation in the north component
        """
        self.add_comp(uy,'uy')
        self.uy=uy
        
    def add_erry(self,erry):
        """
        Adds the uncertainty in the deformation for the north component
        
        Parameters:
            erry (array): uncertainty in the north component
        """
        self.add_errcomp(erry,'uy')
        self.erry=erry
        
    def add_uz(self,uz):
        """
        Adds the deformation in the vertical component
        
        Parameters:
            uz (array): deformation in the vertical component
        """
        self.add_comp(uz,'uz')
        self.uz=uz
        
    def add_errz(self,errz):
        """
        Adds the uncertainty in the deformation for the vertical component
        
        Parameters:
            errz (array): uncertainty in the vertical component
        """
        self.add_errcomp(errz,'uz')
        self.errz=errz
        
    def get_index(self,name):
        """
        Defines the order for the components in the data array
        
        Returns:
            index (int): component's index in the data array
        """
        if name=='ux':
            return 0
        elif name=='uy':
            return 1
        elif name=='uz':
            return 2
    
    def add_data(self,ux,uy,uz):
        """
        Adds the three components at the same time
        
        Returns:
            ux (array): deformation in the east component
            uy (array): deformation in the north component
            uz (array): deformation in the vertical component
        """
        self.add_ux(ux)
        self.add_uy(uy)
        self.add_uz(uz)
        super().add_data(np.concatenate((ux,uy,uz)))
    
    def add_err(self,errx,erry,errz):
        """
        Adds uncertainties for the three components at the same time
        
        Returns:
            errx (array): uncertainty in the east component
            erry (array): uncertainty in the north component
            errz (array): uncertainty in the vertical component
        """
        self.add_errx(errx)
        self.add_erry(erry)
        self.add_errz(errz)
        super().add_err(np.concatenate((errx,erry,errz)))
    
    def importcsv(self,csvfile,ori=None):
        """
        Imports csv file into a gnss object
        
        Parameters:
            csvfile (str): Path to the csv file
            ori (array): lon/lat for the origin coordinate in the projection
        """
        names,lons,lats,uxs,uys,uzs,euxs,euys,euzs=util.read_gnss_csv(csvfile)
        
        self.add_names(np.array(names))
        
        self.add_lls(lons,lats,ori)
        
        self.add_ux(uxs)
        self.add_uy(uys)
        self.add_uz(uzs)
        
        
        self.add_errx(euxs)
        self.add_erry(euys)
        self.add_errz(euzs)
    
    def from_model3d(self,func,unravel=True):
        """
        Uses the function from the forward model to compute the deformation in the three components
        
        Parameters:
            func: 3d displacements forward model function
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
            
        Returns:
            model (array): array containing the components
        """
        ux,uy,uz=func(self.xs,self.ys)
            
        model=()
        
        if isinstance(self.ux,(list,np.ndarray)):
            model=(*model,ux)
        if isinstance(self.uy,(list,np.ndarray)):
            model=(*model,uy)
        if isinstance(self.uz,(list,np.ndarray)):
            model=(*model,uz)
        
        return model


from . import Data
from vmod import util
import numpy as np

class Insar(Data):
    """
    Class that represents the InSAR datatype

    Attributes:
        az: azimuth clockwise angle starting from north (degrees)
        inc: incidence angle where 0 is completely vertical and 90 horizontal (degrees)
        los: line of sight deformation (m)
    """
    def __init__(self):
        self.az=None
        self.inc=None
        self.los=None
        super().__init__()

    def add_az(self,az):
        """
        Adds the azimuth angle(s)
        
        Parameters:
            az: azimuth angle clockwise from north (degrees)
        """
        self.assert_size(az,'az')
        self.az=az
    
    def add_inc(self,inc):
        """
        Adds the incidence angle(s)
        
        Parameters:
            inc: incidence angle(s) with 0 being completely vertical and 90 completely horizontal (degrees)
        """
        self.assert_size(inc,'inc')
        self.inc=inc
        
    def add_vecs(self,az,inc):
        """
        Adds incidence and azimuth angle(s)
        
        Parameters:
            az: azimuth angle clockwise from north (degrees)
            inc: incidence angle(s) with 0 being completely vertical and 90 completely horizontal (degrees)
        """
        self.add_az(az)
        self.add_inc(inc)
        
    def add_los(self,los):
        """
        Adds the line of sight component to the dataset
        
        Parameters:
            los (array): line of sight deformation
        """
        self.add_comp(los,'los')
        self.los=los
        
    def add_ref(self,ref):
        """
        Assigns a reference pixel as reference for the InSAR deformation map
        
        Parameters:
            ref (str): name of the pixel in coordinates
        """
        if self.xs is None or self.ys is None or self.los is None:
            raise Exception('You cannot add a reference to an incomplete dataset')
        else:
            names=np.array([str(self.xs[i])+','+str(self.ys[i]) for i in range(len(self.xs))])
            posmin=np.argmin((self.xs-ref[0])**2+(self.ys-ref[1])**2)
            self.utmz=[ref[2],ref[3],ref[4]]
            if self.names is None:
                self.add_names(names)
            super().add_ref(names[posmin])
    
    def importcsv(self,csvfile,ori=None):
        """
        Imports a csv file into an Insar object
        
        Parameters:
            csvfile (str): path to the csv file
            ori (array): lon/lat for the origin coordinate in the projection
        """
        lons,lats,azs,lks,los,elos,ref=util.read_insar_csv(csvfile)
        
        if np.abs(np.nanmean(azs))>2*np.pi:
            azs=np.radians(azs)
            lks=np.radians(lks)
        self.add_vecs(azs,lks)
        self.add_lls(lons,lats,ori)
        self.add_los(los)
        self.add_err(elos)
        if ref is not None:
            if len(ref)==2:
                orix,oriy,z1sor,z2sor=util.ll2utm([ori[0]],[ori[1]])
                refx,refy,z1sor,z2sor=util.ll2utm([ref[0]],[ref[1]],z1=z1sor,z2=z2sor)
                ref=[refx[0]-orix[0],refy[0]-oriy[0],orix[0],oriy[0],str(z1sor)+str(z2sor)]
            self.add_ref(ref)
        
    def from_model3d(self,func,unravel=True):
        """
        Projects 3d displacements into the line of sight
        
        Parameters:
            func: 3d displacements forward model function
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
        """
        ux,uy,uz=func(self.xs,self.ys)
        assert isinstance(self.inc,(list,np.ndarray)) and isinstance(self.az,(list,np.ndarray)), 'The look angles have not been defined'
        los=ux*np.sin(self.inc)*np.cos(self.az)-uy*np.sin(self.inc)*np.sin(self.az)-uz*np.cos(self.inc)
        los_ref=self.reference_dataset(los,unravel)
        return (-los_ref,)
        


import numpy as np
from .. import util
from . import Source
from .okada import Okada

class RFault(Source):
    """
    A class used to represent a Ring Fault model

    ...

    Attributes
    ----------
    data : Data
        Data object used as input
    parameters : array
        names for the parameters in the model
    ori : int
        orientation for the slip patches in the caldera ring fault, ori=1 for caldera uplift, ori=-1 for caldera collapse (default 1)
    lw : int
        number of patches in the vertical orientation
    segs : int
        number of horizontal segments (default 6)
    """
    def __init__(self, data, ori=None, segs=None, lw=None):
        if lw==None:
            self.lw=1
        else:
            self.lw=lw
            
        if ori==None:
            self.ori=1
        else:
            self.ori=ori
            
        if segs==None:
            self.segs=6
        else:
            self.segs=segs
        
        super().__init__(data)
    
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "RFault"

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("RFault")
        print("\tx = %f" % x[0])
        print("\ty = %f" % x[1])
        print("\td = %f" % x[2])
        print("\ts= %f" % x[3])
        print("\to= %f" % x[4])
        print("\ta= %f" % x[5])
        print("\tb= %f" % x[6])
        print("\tW= %f" % x[7])
        print("\tstrike= %f" % x[8])
        
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","slip","opening","smajor","sminor","width","strike")

    # =====================
    # Forward Models
    # =====================
    def get_parameters(self,xcen,ycen,a,b,strike):
        """
        Function defining parameters for the patches that compose the ring fault.
        
        Parameters:
           xcen (float) : x coordinate for the center of the caldera in meters.
           ycen (float) : y coordinate for the center of the caldera in meters.
           a (float) : semimajor axis for the caldera.
           b (float) : semiminor axis for the caldera.
           strike (deg) : azimuth for the caldera clockwise is positive from North.
        
        Returns:
            params (array) : Matrix where each column has the parameters for an Okada patch.
        """
        angles=np.linspace(0,360,self.segs+1)
        params=np.ones((len(angles)-1,8))
        for i,angle in enumerate(angles[1::]):
            pangle=angles[i]
            #print(pangle,angle)
            xc1 = a*np.cos(np.radians(pangle))*np.cos(np.radians(90-strike)) -\
                 b*np.sin(np.radians(pangle))*np.sin(np.radians(90-strike)) + xcen
            yc1 = a*np.cos(np.radians(pangle))*np.sin(np.radians(90-strike)) +\
                 b*np.sin(np.radians(pangle))*np.cos(np.radians(90-strike)) + ycen
            xc2 = a*np.cos(np.radians(angle))*np.cos(np.radians(90-strike)) -\
                 b*np.sin(np.radians(angle))*np.sin(np.radians(90-strike)) + xcen
            yc2 = a*np.cos(np.radians(angle))*np.sin(np.radians(90-strike)) +\
                 b*np.sin(np.radians(angle))*np.cos(np.radians(90-strike)) + ycen
            sxc = np.mean([xc1,xc2])
            syc = np.mean([yc1,yc2])
            slength=np.sqrt((xc2-xc1)**2+(yc2-yc1)**2)
            sstrike=90-np.degrees(np.arctan2(yc2-yc1,xc2-xc1))
            if sstrike<0:
                sstrike=360+sstrike
            params[i,0]=xc1
            params[i,1]=yc1
            params[i,2]=xc2
            params[i,3]=yc2
            params[i,4]=sxc
            params[i,5]=syc
            params[i,6]=slength
            params[i,7]=sstrike
        return params
    
    def get_greens(self,x,y, xcen, ycen, depth, a, b, width, strike):
        """
        Green functions for the ring fault.
        
        Parameters:
           xcen (float) : x coordinate for the center of the caldera in meters.
           ycen (float) : y coordinate for the center of the caldera in meters.
           depth (float) : depth for the Okada patches.
           width (float) : vertical width for the Okada patches.
           a (float) : semimajor axis for the caldera.
           b (float) : semiminor axis for the caldera.
           strike (deg) : azimuth for the caldera clockwise is positive from North.
        
        Returns:
            G (array) : G-Matrix containing the green functions.
        """
        if self.ori==1:
            rake=90.0
        else:
            rake=-90.0
        
        ok = Okada(None)
        G=np.zeros((len(x)*3,self.segs*self.lw*2))
        params=self.get_parameters(xcen,ycen,a,b,strike)
        for i in range(params.shape[0]):
            xc1=params[i,0]
            yc1=params[i,1]
            xc2=params[i,2]
            yc2=params[i,3]
            sxc=params[i,4]
            syc=params[i,5]
            slength=params[i,6]
            sstrike=params[i,7]
            
            ok.set_type('slip')
            ux,uy,uz=ok.model(x,y, sxc, syc, depth, slength, width, 1.0, sstrike, 90.0, rake)
            G[0:len(x),i]=ux
            G[len(x):2*len(x),i]=uy
            G[2*len(x):3*len(x),i]=uz
            
            ok.set_type('open')
            ux,uy,uz=ok.model(x,y, sxc, syc, depth, slength, width, 1.0, sstrike,90.0)
            G[0:len(x),self.segs*self.lw+i]=ux
            G[len(x):2*len(x),self.segs*self.lw+i]=uy
            G[2*len(x):3*len(x),self.segs*self.lw+i]=uz
        
        return G #returns tuple
        
    def model(self,x,y, xcen, ycen, depth, slip, opening, a, b, width, strike):       
        """
        Green functions for the ring fault.
        
        Parameters:
           xcen (float) : x coordinate for the center of the caldera in meters.
           ycen (float) : y coordinate for the center of the caldera in meters.
           depth (float) : depth for the Okada patches.
           slip (float) : slip in the Okada patches in meters.
           opening (float) : opening in the Okada patches in meters.
           a (float) : semimajor axis for the caldera.
           b (float) : semiminor axis for the caldera.
           width (float) : vertical width for the Okada patches.
           strike (deg) : azimuth for the caldera clockwise is positive from North.
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        if isinstance(slip,(list,np.ndarray)):
            if not len(slip)==self.segs*self.lw:
                print('The slip elements do not agree with patches. Taking the mean...')
                slip=np.ones((len(self.segs*self.lw),))*np.mean(slip)
        else:
            slip=np.ones((self.segs*self.lw,))*slip
                
        if isinstance(opening,(list,np.ndarray)):
            if not len(opening)==self.segs*self.lw:
                print('The opening elements do not agree with patches. Taking the mean...')
                opening=np.ones((len(self.segs*self.lw),))*np.mean(opening)
        else:
            opening=np.ones((self.segs*self.lw,))*opening
        
        G=self.get_greens(x,y, xcen, ycen, depth, a, b, width, strike)
        model=np.concatenate((slip,opening))
        
        data=G@model
        
        ux=data[0:len(x)]
        uy=data[len(x):2*len(x)]
        uz=data[2*len(x):3*len(x)]
        
        return ux,uy,uz



import numpy as np
import scipy
from ..data import Data

class Source:
    """
    Base class for analytical magmatic source models. Implements
    common functions required from all child classes.

    Attributes
        data (Data): data object
        x0 (array): initial guess for parameter values
        offsets (boolean): compute offsets as parameters for each component in the data object
        low_bounds (array): lower limit for the parameter values
        high_bounds (array): upper limit for the parameter values
    """
    def __init__(self, data):
        self.data        = data
        self.x0          = None
        self.offsets     = False
        self.parameters  = None
        self.reg         = False
        self.low_bounds  = []
        self.high_bounds = []
        
    def add_offsets(self):
        """
        Add offsets as parameters for each component in the dara object
        """
        if self.data.ts is None:
            self.offsets=True
            self.get_parnames()
        else:
            raise Exception('The data has a time dependency')
        
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=1100000
        burnin=100000
        thin=1000
        
        return steps,burnin,thin
    
    def get_parnames(self):
        """
        Function that add offsets to the list of parameters.
        """
        self.set_parnames()
        if self.offsets:
            for i,c in enumerate(self.data.comps):
                self.parameters=(*self.parameters,'offset'+str(i))
        return self.parameters
    
    def get_num_params(self):
        """
        Function that give the number of parameters.
        
        Returns:
            size (int): length of the parameters.
        """
        return len(self.parameters)

    def set_x0(self, x0):
        """
        Function that sets the initial guess for the model.
        
        Parameters:
            x0 (list): list of values.
        """
        self.get_parnames()
        self.x0 = x0

    def set_bounds(self, low_bounds, high_bounds):
        """
        Function that sets the low and upper bounds for the parameters.
        
        Parameters:
            low_bounds (list): lower bounds for the parameters.
            high_bounds (list): upper bounds for the parameters.
        """
        self.get_parnames()
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

    def get_xs(self):
        """
        Function that gives the data points positions in east.
        
        Returns:
            xs (list): positions in east.
        """
        return self.data.xs

    def get_ys(self):
        """
        Function that gives the data points positions in north.
        
        Returns:
            ys (list): positions in north.
        """
        return self.data.ys

    def get_ts(self):
        """
        Function that gives the times for the observations.
        
        Returns:
            ts (list): times for the observations.
        """
        return self.data.ts
    
    def get_zs(self):
        """
        Function that gives the data points positions in vertical.
        
        Returns:
            zs (list): positions in vertical.
        """
        return self.data.zs
    
    def get_orders(self):
        """
        Function that gives the orders for the parameters value.
        
        Returns:
            orders (list): orders for the parameters.
        """
        orders=[]
        for i in range(len(self.low_bounds)):
            order=int(np.log10(np.max([np.abs(self.low_bounds[i]),np.abs(self.high_bounds[i])])))-1
            orders.append(10**order)
        orders=np.array(orders)
        return orders
    
    def strain(self,x,y,args):
        """
        Function that computes stresses in the horizontal plane.
        Parameters:
            x: x-coordinate (m)
            y: y-coordinate (m)
            args: parameters for the model
        Returns:
            sxx (list): normal strain in the x direction.
            syy (list): normal strain in the y direction.
            sxy (list): horizontal shear strain.
        """
        hx=0.001*np.abs(np.max(x)-np.min(x))
        hy=0.001*np.abs(np.max(y)-np.min(y))
        if hx==0:
            h=hy
        elif hy==0:
            h=hx
        elif hx<hy:
            h=hx
        elif hy<hx:
            h=hy
            
        u,v,w=self.model(x, y, *args)
        
        upx,vpx,wpx=self.model(x+h, y, *args)
        umx,vmx,wmx=self.model(x-h, y, *args)
        dudx=0.5*(upx-umx)/h
        dvdx=0.5*(vpx-vmx)/h
        dwdx=0.5*(wpx-wmx)/h
        
        upy,vpy,wpy=self.model(x, y+h, *args)
        umy,vmy,wmy=self.model(x, y-h, *args)
        dudy=0.5*(upy-umy)/h
        dvdy=0.5*(vpy-vmy)/h
        dwdy=0.5*(wpy-wmy)/h
        
        sxx=2*dudx
        syy=2*dvdy
        sxy=(dudy+dvdx)
        
        return sxx,syy,sxy
        
    def stress(self, x, y, z, args):
        """
        Function that computes stresses in the horizontal plane.
        
        Parameters:
            x: x-coordinate (m)
            y: y-coordinate (m)
            args: parameters for the model
        
        Returns:
            sxx (list): normal stress in the x direction (Pa).
            syy (list): normal stress in the y direction (Pa).
            szz (list): normal stress in the z direction (Pa).
            sxy (list): shear stress in the xy direction (Pa).
            sxz (list): shear stress in the xz direction (Pa).
            syz (list): shear stress in the yz direction (Pa).
        """
        hx=0.001*np.abs(np.max(x)-np.min(x))
        hy=0.001*np.abs(np.max(y)-np.min(y))
        if hx==0:
            h=hy
        elif hy==0:
            h=hx
        elif hx<hy:
            h=hx
        elif hy<hx:
            h=hy
            
        u,v,w=self.model_depth(x, y, z, *args)
        
        upx,vpx,wpx=self.model_depth(x+h, y, z, *args)
        umx,vmx,wmx=self.model_depth(x-h, y, z, *args)
        dudx=0.5*(upx-umx)/h
        dvdx=0.5*(vpx-vmx)/h
        dwdx=0.5*(wpx-wmx)/h
        
        upy,vpy,wpy=self.model_depth(x, y+h, z, *args)
        umy,vmy,wmy=self.model_depth(x, y-h, z, *args)
        dudy=0.5*(upy-umy)/h
        dvdy=0.5*(vpy-vmy)/h
        dwdy=0.5*(wpy-wmy)/h
        
        double=False
        if isinstance(z,float):
            if z==0:
                double=True
            else:
                double==False
        elif len(z[z==0])>0:
            double=True
            
        if double:
            upz,vpz,wpz=self.model_depth(x, y, z+2*h, *args)
            umz,vmz,wmz=self.model_depth(x, y, z, *args)
        else:
            upz,vpz,wpz=self.model_depth(x, y, z+h, *args)
            umz,vmz,wmz=self.model_depth(x, y, z-h, *args)
        dudz=0.5*(upz-umz)/h
        dvdz=0.5*(vpz-vmz)/h
        dwdz=0.5*(wpz-wmz)/h
        
        
        nu=args[-1]
        mu=args[-2]
        
        sxx=2*(1+nu)*dudx*mu
        syy=2*(1+nu)*dvdy*mu
        szz=2*(1+nu)*dwdz*mu
        sxy=(1+nu)*(dudy+dvdx)*mu
        sxz=(1+nu)*(dudz+dwdx)*mu
        syz=(1+nu)*(dvdz+dwdy)*mu
        
        return sxx,syy,szz,sxy,sxz,syz
    
    def forward(self,args,unravel=True):
        """
        Function that computes the forward model.
        
        Parameters:
            args: parameters for the model
        
        Returns:
            output (list): output in certain datatype according to the data object.
        """
        self.get_parnames()
        if self.offsets:
            offsets=args[-len(self.data.comps)::]
            args=args[0:len(args)-len(self.data.comps)]
        else:
            offsets=None
        if self.data.ts is None:
            if self.data.__class__.__name__=='Tilt' and 'model_tilt' in dir(self):
                func_tilt=lambda x,y: self.model_tilt(x,y,*args)
                func_tilt.__name__ = 'func_tilt'
                return self.data.from_model(func_tilt,offsets,unravel)
            elif 'model' in dir(self):
                func=lambda x,y: self.model(x,y,*args)
                return self.data.from_model(func,offsets,unravel)
            else:
                raise Exception('The source does not have a time-independent model defined')
        else:
            if self.data.__class__.__name__=='Tilt' and 'model_tilt_t' in dir(self):
                func_tilt_time=lambda x,y,t: self.model_tilt_t(x,y,t,*args)
                func_tilt.time.__name__ = 'func_tilt_time'
                return self.data.from_model(func_tilt_time,offsets,unravel)
            elif 'model_t' in dir(self):
                func_time=lambda x,y,t: self.model_t(x,y,t,*args)
                func_time.__name__ = 'func_time'
                return self.data.from_model(func_time,offsets,unravel)
            else:
                raise Exception('The source does not have a time-dependent model defined')
    


import numpy as np
from .. import util
from . import Source

class Vshell(Source):
    """
    Class that represents a pressurized sphere within a viscoelastic shell within an elastic medium (Segall, 2010).

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Vshell"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=1100000
        burnin=10000
        thin=1000
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Vshell")
        print("\tx = %f" % x[0])
        print("\ty = %f" % x[1])
        print("\td = %f" % x[2])
        print("\tradius1 = %f" % x[3])
        print("\tradius2 = %f" % x[4])
        print("\tdP= %f" % x[5])
        print("\ttau= %f" % x[6])
    
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","radius","dP","tau")
        
    def model(self, x, y, xcen, ycen, d, rad, dP):
        """
        Initial 3d displacement field on surface for viscoelastic shell (Segall, 2010)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dP: change in pressure (Pa)
            
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        return model_t(x,y,0, xcen, ycen, d, rad, dP)
        
    def model_t(self, x, y, t, xcen, ycen, d, rad1, rad2, dP, tau, nu=0.25, mu=4e9):
        """
        3d displacement field on surface for viscoelastic shell (Segall, 2010)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            t: input time (s)
            xcen: x-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dP: change in pressure (Pa)
            
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        x = x - xcen
        y = y - ycen
        
        coeff=(1-nu)*dP*rad1**3/(mu*d**2)
        
        th, r = util.cart2pol(x,y)
        
        uz=coeff*(np.exp(-t/tau)+((rad2/rad1)**3)*(1-np.exp(-t/tau)))/(1+r**2)**1.5

        ur=uz*r
        
        ux, uy = util.pol2cart(th, ur)
        
        return ux,uy,uz


from .gnss import Gnss

class Level(Gnss):
    """
    Class that represents a a levelling dataset. It uses the Gnss class
    """
    def add_data(self,deltas):
        """
        Adds the vertical components
        
        Parameters:
            deltas (array): changes in the vertical (m)
        """
        self.add_uz(deltas)
    
    def add_err(self,err):
        """
        Adds the uncertainties in the vertical components
        
        Parameters:
            err (array): uncertainties in the changes in the vertical (m)
        """
        self.add_errz(err)


import numpy as np
from scipy import special
from . import Source
from .. import util
from scipy import integrate

class Wellsite(Source):
    """
    Class that represents a pressurized well following the implementation from Wangen (2018).

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Wellsite"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=3100000
        burnin=10000
        thin=1000
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Wellsite")
        print("\txcen = %f" % x[0])
        print("\tycen = %f" % x[1])
        print("\twidth = %f" % x[2])
        print("\tpressure = %f" % x[3])
        print("\tdiffusivity= %f" % x[4])
        print("\tYoung Modulus= %f" % x[5])
        
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","width","pressure","diffu","Young modulus")
    
    # =====================
    # Forward Models
    # =====================
    def pressure(self,x,P,diffu,t):
        """
        The function calculates the pressure as a function of time.
        
        Parameters:
           x (float): distance to the injection well (m)
           P (float): initial pressure (Pa)
           diffu (float): diffusivity coefficient (m^2/s)
           t: input time (s)
           
        Results:
            pressure: pressure at given time (Pa)
        """
        return P*special.erfc(x/(2*np.sqrt(diffu*t)))
    
    def dpressure(self,x,an,fn,k):
        return fn*an*np.cos(k*x)
    
    def model_t(self, x, y, t, xcen, ycen, depth, width, P, diffu, E=5, nu=0.2):
       
        """
        The function 
        
        Parameters:
            xcen: y-offset of point source epicenter (m): []
            ycen: y-offset of point source epicenter (m): []
            depth: Reservoir depth (m): 
            width: reservoir width (m)
            P: Initial pressure (MPa): [5-50]
            diffu: Diffusivity (m^2/s): [10^-4-10^-1]
            E: Young's modulus (GPa): [5-15]
            nu: Poisson ratio
        
        Returns:
            ux: the model does not compute deformation in the x-axis
            uy: the model does not compute deformation in the y-axis
            uz: deformation in the vertical (m)
        """
        alpha=1
        
        mu=E/(2*(1+nu))
        
        Lambda = (nu*E)/((1+nu)*(1-2*nu))
        
        h1=width #[10-100]
        h2=depth #[50-300]
        
        # center coordinate grid on point source
        x = x - xcen
        y = y - ycen

        # convert to surface cylindrical coordinates
        th, r = util.cart2pol(x,y)
        
        H = h1+h2
        L =np.nanmax(r)               #m, minimum domain size, >>2*np.pi*H
        N=100

        
        tunique=list(set(t))
        uz=x*np.nan
        for tt in tunique:
            ks=[]
            wlns=[]
            fns=[]
            pus =[]
            ws =[]
            rt=r[t==tt]
            
            rsimp=np.linspace(0,np.max(rt),1000)

            px_simp=self.pressure(rsimp,P,diffu,tt)
            #px_simp=lambda r: self.pressure(r,P,diffu,tt)
            ans=[]
            px_Fs=rt*0

            for n in range(0,N):
                if n == 0:
                    k=0
                    fn = 1/2
                    fns.append(fn)
                    #fun=lambda r: self.pressure(r,P,diffu,tt)*np.cos(n*np.pi*r/L)
                    #an = ((2/L)*integrate.quad_vec(fun,0,L)[0]) #Fourier coeff an
                    an = ((2/L)*np.sum(px_simp*np.cos(n*np.pi*rsimp/L)*np.abs(rsimp[1]-rsimp[0])))
                else:    
                    wln = (2*L)/n        #wavelength n
                    k = (2*np.pi)/wln    #wavenumber n
                    fn = (1/(k*h1))*((np.cosh(k*h2)*np.tanh((k*h1)+(k*h2)))-np.sinh(k*h2)) #dimensionless amplitude dep. on h1 & h2

                    #Fourier coeff an
                    #fun=lambda r: self.pressure(r,P,diffu,tt)*np.cos(n*np.pi*r/L)
                    #an = ((2/L)*integrate.quad_vec(fun,0,L)[0]) #Fourier coeff an
                    #an = ((2/L)*integrate.cumtrapz(px_simp*np.cos(n*np.pi*rt/L)))
                    an = ((2/L)*np.sum(px_simp*np.cos(n*np.pi*rsimp/L)*np.abs(rsimp[1]-rsimp[0])))
                    
                px_Fs+=self.dpressure(rt,an,fn,k)

            uz[t==tt] = (alpha*h1*px_Fs)/(Lambda + (2*mu))
            fns[0]=1

        ux=0*uz
        uy=0*uz
        
        return ux, uy, uz


from . import Data
import numpy as np
import copy

class Joint(Data):
    """
    Class that represents a joint dataset composed by one or many different datasets

    Attributes:
        datasets (array): array with multiple data objects
        comps (array): array containing all the components in the data objects
    """
    def __init__(self):
        self.datasets=None
        super().__init__()
    
    def add_dataset(self,data,wt=1.0):
        """
        Adds a data object to the joint dataset
        
        Parameters:
            data (Data): data object that represents a dataset
            wt (float): relative weight for the dataset
        """
        dataset=copy.deepcopy(data)
        if isinstance(dataset,Data):
            if dataset.err is not None:
                dataset.err=dataset.err/(np.min(dataset.err)*wt)
            if self.datasets is None:
                if dataset.data is None:
                    raise Exception('There are not data points in this dataset')
                else:
                    self.datasets=[dataset]
                    self.comps=copy.deepcopy(dataset.comps)
            else:
                self.datasets.append(dataset)
                self.comps+=dataset.comps
                
                
    def ref_possible(self):
        """
        Overrides the function to specify that the joint dataset cannot take references
        """
        return False
    
    def get_data(self):
        """
        Concatenates and gives the data in the datasets
        
        Returns:
            data (array): deformation data
        """
        data=[]
        for dataset in self.datasets:
            data+=dataset.get_data().tolist()
        return np.array(data)
    
    def get_errors(self):
        """
        Concatenates and gives the uncertainties in the datasets
        
        Returns:
            errors (array): uncertainties deformation data
        """
        errors=[]
        for dataset in self.datasets:
            errors+=dataset.get_errors().tolist()
        return np.array(errors)
    
    def from_model(self,func,offsets=None,unravel=True):
        """
        Uses the function from_model on each dataset to compute the forward model
        
        Parameters:
            func: forward model function
            offsets: None or array that contains the offset for each component
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
            
        Returns:
            ux: modeled deformation in the x-axis
            uy: modeled deformation in the y-axis
            uz: modeled deformation in the vertical
        """
        if self.datasets is None:
            raise Exception('No datasets included')
        if offsets is None:
            offsets=np.array([0 for i in self.comps])
        data=[]
        ini=0
        for dataset in self.datasets:
            ncomps=ini+len(dataset.comps)
            datat=dataset.from_model(func,offsets[ini:ncomps],unravel)
            data+=datat.tolist()
            ini=len(dataset.comps)
        return np.array(data)
            